import json
from .card import Card
from .deck import Deck


class Rectangle:
    def __init__(self, **kwargs):
        self.card_id = kwargs.get("CardID", 0)
        self.card_code = kwargs.get("CardCode", None)
        self.top_left_x = kwargs.get("TopLeftX", 0)
        self.top_left_y = kwargs.get("TopLeftY", 0)
        self.width = kwargs.get("Width", 0)
        self.height = kwargs.get("Height", 0)
        self.isLocalPlayer = kwargs.get("LocalPlayer")
        self.card = Card(self.card_code)

    def __repr__(self):
        return f"Rectangle(Card: {self.card_code})"


class Screen:
    def __init__(self, **kwargs):
        self.width = kwargs.get("ScreenWidth", 0)
        self.height = kwargs.get("ScreenHeight", 0)


class GameFrame:
    def __init__(self, **kwargs):
        self.player = kwargs.get("PlayerName", "The Man With No Name")
        self.opponent = kwargs.get("OpponentName", "The Man With No Name")
        self.game_state = kwargs.get("GameState", "Menus")
        self.screen = Screen(**kwargs.get("Screen", None))
        self._rectangles = kwargs.get("Rectangles", [])
        self.rectangles = self.parse_rectangles()

    def parse_rectangles(self):
        rects = [
            rect for rect in self._rectangles if rect["CardCode"] != "face"
        ]
        return list(map(lambda x: Rectangle(**x), rects))

    @property
    def player_rects(self):
        return filter(lambda x: x.isLocalPlayer, self.rectangles)

    @property
    def opponent_rects(self):
        return filter(lambda x: not x.isLocalPlayer, self.rectangles)


class Game:
    def __init__(self, player, opponent, screen, player_deck):
        self.player = player
        self.opponent = opponent
        self.screen = screen
        self.player_cards_used = Deck()
        self.opponent_cards_used = Deck()
        self.initial_player_deck = self.current_player_deck = player_deck
        self.result = None

    def process_frame(self, frame):
        for rect in frame.player_rects:
            if rect.card_id not in [
                    card.id for card in self.player_cards_used.cards
            ]:
                current_card = Card(CardID=rect.card_id,
                                    CardCode=rect.card_code)
                self.player_cards_used.add_card(current_card)
                #remove card from current player deck

        for rect in frame.opponent_rects:
            if rect.card_id not in [
                    card.id for card in self.opponent_cards_used.cards
            ]:
                current_card = Card(CardID=rect.card_id,
                                    CardCode=rect.card_code)
                self.opponent_cards_used.add_card(current_card)


class ExpeditionState:
    def __init__(self, **kwargs):
        self.is_active = kwargs.get("IsActive", False)
        self.state = kwargs.get("State", "Inactive")
        self.record = kwargs.get("Record", [])
        self.draft_picks = kwargs.get("DraftPicks", [])
        self.deck = kwargs.get("Deck")  # TODO: convert to Deck instance
        self.games_played = kwargs.get("Games", 0)
        self.wins = kwargs.get("Wins", 0)
        self.losses = kwargs.get("Losses", 0)

    def __repr__(self):
        return f"Expedition(State: {self.state}, Games Played: {self.games_played})"


name = "twisted_fate"
__version__ = "0.1.4"
from .api_wrapper.deck import Deck
from .api_wrapper.card import Card
from .api_wrapper.active_game import Game, GameFrame, Screen
from .api_wrapper.client_api import LoRClient
# from .deck_coder.deckCoder import DeckCode



import os
import requests
import logging
import json
from .deck import Deck
from .active_game import GameFrame, ExpeditionState

logger = logging.getLogger("TwistedFateLib")
logger.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(name)s :: %(levelname)s | %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)


class GameStatus:
    def __init__(self, game_id, is_winner):
        self.game_id = game_id
        self.is_winner = is_winner

    @property
    def result(self):
        if self.game_id == -1:
            return "No games played yet this session"
        if self.is_winner: return "Win"
        else: return "Loss"

    def serialize(self, to_dict=False):
        data = {"game_id": self.game_id, "result": self.result}
        return data if to_dict else json.dumps(data)


class LoRClient:

    baseurl = "http://localhost:"

    def __init__(self, api_key, port=21337):
        self.api_key = api_key
        self.port = "port"
        self.baseurl = f"{LoRClient.baseurl}{port}"

    def get_endpoint(self, endpoint):
        url = f"{self.baseurl}/{endpoint}"
        logger.info(f"Getting {endpoint}")
        try:
            response = requests.get(url)
        except:
            logger.error("Cannot connect to LoR Client")
            return
        status = f"{response.status_code} - {response.ok}"
        logger.info(f"Endpoint: {endpoint} response {status}")
        return response.json()

    def current_decklist(self) -> Deck:
        r = self.get_endpoint("static-decklist")
        deck = Deck(**r)
        return deck

    def card_positions(self) -> GameFrame:
        r = self.get_endpoint("positional-rectangles")
        if r:
            logging.debug(r)
            frame = GameFrame(**r)
            return frame
        # return r

    def game_status(self) -> GameStatus:
        r = self.get_endpoint("game-result")
        if r:
            game = GameStatus(r['GameID'], r["LocalPlayerWon"])
            return game

    def expeditions_state(self) -> ExpeditionState:
        r = self.get_endpoint("expeditions-state")
        if r:
            exp = ExpeditionState(**r)
            return exp



# Credit to @stelar7, this python implementation is a port of his original javascript implementation

from base64 import b32decode, b32encode

MAX_KNOWN_VERSION = 2


class Base32:
    @staticmethod
    def decode(b32string):
        s = Base32.add_padding(b32string)
        return b32decode(s)

    @staticmethod
    def encode(bytes_list):
        encoded_string = b32encode(bytes(bytes_list))
        encoded_string = str(encoded_string, "utf-8").strip("=")
        return encoded_string

    # python's builtin decoder is strict with padding
    @staticmethod
    def add_padding(b32String):
        length = len(b32String)
        padding = 0

        if (length % 8 > 0):
            padding = 8 - (length % 8)
            b32String += "=" * padding
        return b32String


faction_code_to_id = {
    "DE": 0,
    "FR": 1,
    "IO": 2,
    "NX": 3,
    "PZ": 4,
    "SI": 5,
    "BW": 6
}
id_to_faction_code = {
    0: "DE",
    1: "FR",
    2: "IO",
    3: "NX",
    4: "PZ",
    5: "SI",
    6: "BW"
}


class DeckCode:
    @staticmethod
    def encode_deck(deck):

        result = [17]
        cards_3 = []
        cards_2 = []
        cards_1 = []
        cards_other = []

        if not DeckCode.is_valid_card_codes_and_count(deck):
            raise Exception("Deck contains invalid card codes")
        else:
            for card, count in deck.items():
                if count == 3:
                    cards_3.append(card)
                elif count == 2:
                    cards_2.append(card)
                elif count == 1:
                    cards_1.append(card)
                else:
                    cards_other.append((card, count))

        grouped_cards_3 = DeckCode.group_by_faction(cards_3)
        grouped_cards_2 = DeckCode.group_by_faction(cards_2)
        grouped_cards_1 = DeckCode.group_by_faction(cards_1)

        grouped_cards_3 = DeckCode.sort_card_groups(grouped_cards_3)
        grouped_cards_2 = DeckCode.sort_card_groups(grouped_cards_2)
        grouped_cards_1 = DeckCode.sort_card_groups(grouped_cards_1)
        cards_other = sorted(cards_other)

        # encode
        result = [*result, *VarIntTransformer.encodeGroupOf(grouped_cards_3)]
        result = [*result, *VarIntTransformer.encodeGroupOf(grouped_cards_2)]
        result = [*result, *VarIntTransformer.encodeGroupOf(grouped_cards_1)]

        result = [*result, *VarIntTransformer.encodeNOfs(cards_other)]

        return Base32.encode(result)

    @staticmethod
    def group_by_faction(card_list):
        new_list = []
        while len(card_list) > 0:
            card = card_list.pop(0)
            set_num, faction, _ = DeckCode.parse_card_code(card)

            faction_set = []
            faction_set.append(card)

            temp_list = card_list[:]
            for card in temp_list:
                current_set_num, current_faction, _ = DeckCode.parse_card_code(
                    card)
                if current_set_num == set_num and current_faction == faction:
                    faction_set.append(card)
                    card_list.remove(card)
            faction_set.sort()
            new_list.append(faction_set)

        return new_list

    @staticmethod
    def sort_card_groups(card_groups):
        s = sorted(card_groups, key=lambda x: len(x))
        return s

    @staticmethod
    def parse_card_code(card_code):
        set = card_code[:2]
        faction = card_code[2:4]
        card_id = card_code[4:]
        return set, faction, card_id

    @staticmethod
    def is_valid_card_codes_and_count(deck):
        for card, count in deck.items():
            code = card
            if len(code) != 7:
                print(f"deck code length is {len(code)}")
                return False

            # check set code is numeric
            for char in code[:2]:
                if not char.isdigit():
                    print(f"code is not digit: {char}")
                    return False

            faction = faction_code_to_id.get(code[2:4], False)
            if not faction:
                print(f"faction code not in faction code list: {code[2:4]}")
                return False

            for char in code[4:]:
                if not char.isdigit():
                    print(f"card number contains non digit: {char}")
                    return False

            if count < 1:
                print("less than one card: {count}")
                return False

        return True

    @staticmethod
    def decode_deck(string):
        result = {}

        byte_list = bytearray(Base32.decode(string))
        # print(type(byte_list))

        _format = byte_list[0] >> 4
        version = byte_list[0] & 0xF
        byte_list = byte_list[1:]

        if version > MAX_KNOWN_VERSION:
            raise ValueError(
                "Please update to the latest version of twisted_fate")

        for i in range(3, 0, -1):
            numGroupOfs = VarIntTransformer.popVarInt(byte_list)

            for __ in range(numGroupOfs):
                numOfsInThisGroup = VarIntTransformer.popVarInt(byte_list)
                setNum = VarIntTransformer.popVarInt(byte_list)
                faction = VarIntTransformer.popVarInt(byte_list)

                for ___ in range(numOfsInThisGroup):
                    card = VarIntTransformer.popVarInt(byte_list)

                    setString = str(setNum).zfill(2)
                    factionString = id_to_faction_code[faction]
                    cardString = str(card).zfill(3)

                    card_code = setString + factionString + cardString
                    result[card_code] = i

        while len(byte_list) > 0:
            fpc = VarIntTransformer.popVarInt(byte_list)
            fps = VarIntTransformer.popVarInt(byte_list)
            fpf = VarIntTransformer.popVarInt(byte_list)
            fpn = VarIntTransformer.popVarInt(byte_list)

            fpss = str(fps).zfill(2)
            fpfs = id_to_faction_code[fpf]
            fpns = str(fpn).zfill(3)

            card_code = fpss + fpfs + fpns
            result[card_code] = fpc

        return result


class VarIntTransformer:
    @staticmethod
    def popVarInt(_bytes: bytes):

        # _bytes = bytearray(_bytes)
        AllButMSB = 0x7F
        JustMSB = 0x80

        result = 0
        current_shift = 0
        bytes_popped = 0

        for i in range(len(_bytes)):
            bytes_popped += 1
            # print(len(_bytes))
            current = _bytes[i] & AllButMSB
            result |= current << current_shift

            if (i & JustMSB) != JustMSB:
                del _bytes[:bytes_popped]
                return result

            current_shift += 7

        raise Exception("Byte array did not contain valid variants.")

    @staticmethod
    def getVarInt(value):
        value = int(value)
        AllButMSB = 0x7F
        JustMSB = 0x80

        buff = [10]
        current_index = 0

        if value == 0:
            return [0]

        while value != 0:
            byte_val = value & AllButMSB
            value >>= 7

            if value != 0:
                byte_val |= JustMSB

            try:
                buff[current_index] = byte_val
            except IndexError:
                buff.append(byte_val)

            current_index += 1

        return buff[:current_index]

    @staticmethod
    def encodeNOfs(nOfs):
        _bytes = []
        for card, count in nOfs:
            _bytes = [*_bytes, *VarIntTransformer.getVarInt(count)]

            setNum, factionCode, cardNum = DeckCode.parse_card_code(card)
            factionNum = faction_code_to_id[factionCode]

            _bytes = [*_bytes, *VarIntTransformer.getVarInt(setNum)]
            _bytes = [*_bytes, *VarIntTransformer.getVarInt(factionNum)]
            _bytes = [*_bytes, *VarIntTransformer.getVarInt(cardNum)]

        return _bytes

    @staticmethod
    def encodeGroupOf(groupOf):
        _bytes = []

        _bytes = [*_bytes, *VarIntTransformer.getVarInt(len(groupOf))]

        for cardList in groupOf:
            _bytes = [*_bytes, *VarIntTransformer.getVarInt(len(cardList))]

            current_card_code = cardList[0]
            current_set_num, current_faction_code, _ = DeckCode.parse_card_code(
                current_card_code)
            current_faction_num = faction_code_to_id[current_faction_code]

            _bytes = [*_bytes, *VarIntTransformer.getVarInt(current_set_num)]
            _bytes = [
                *_bytes, *VarIntTransformer.getVarInt(current_faction_num)
            ]

            for card in cardList:
                sequenceNumber = card[4:]
                _bytes = [
                    *_bytes, *VarIntTransformer.getVarInt(sequenceNumber)
                ]

        return _bytes



from pathlib import Path
import os
import json
from .card import Card
from ..deck_coder.deckCoder import DeckCode
from .utils import read_json_file, get_lor_globals
from collections import Counter

try:
    globals_file = Path("./data/data/globals-en_us.json")
    data_globals = read_json_file(globals_file)
except:
    data_globals = get_lor_globals()


class Deck:
    def __init__(self, **kwargs):
        self._cards = kwargs.get("CardsInDeck", kwargs.get("cards", []))
        self.deck_code = kwargs.get("DeckCode", None)
        self.cards = []
        self.deck_id = kwargs.get("deck_id", None)
        self.wins = kwargs.get("wins", 0)
        self.losses = kwargs.get("losses", 0)

        if self.deck_code:
            self.decode(self.deck_code, instance=self, in_place=True)

        if self._cards:
            for card, amount in self._cards.items():
                self.cards.append((Card(CardCode=card, count=amount)))

        if not self.cards and self.deck_code:
            pass  # TODO: generate cards list from deck code

    def encode(self):
        if not self.deck_code:
            self.deck_code = DeckCode.encode_deck(self._cards)
        return self

    @classmethod
    def decode(cls, deck_code, instance=None, in_place=False):
        cards = DeckCode.decode_deck(deck_code)
        if in_place:
            instance._cards = cards
            return None
        return cls(CardsInDeck=cards)

    def to_deck_code(self):
        if not self._cards:
            raise ValueError("Deck is empty")
        if not self.deck_code:
            self.encode()
        return self.deck_code

    def regions(self):
        card_regions = [card.region for card in self.cards]
        region_count = Counter(card_regions)
        card_regions = region_count.most_common(2)
        regions = [x[0] for x in card_regions]
        return regions

    def champions(self):
        champs = filter(lambda x: x.isChampion, self.cards)
        return [champ.name for champ in champs]

    def serialize(self, as_dict=False):
        deck = {}
        deck["cards"] = [c.serialize(as_dict=True) for c in self.cards]
        deck["id"] = self.deck_id
        deck["champions"] = self.champions()
        deck["regions"] = self.regions()
        deck["code"] = self.to_deck_code()
        deck["wins"] = self.wins
        deck["losses"] = self.losses
        if as_dict:
            return deck
        return json.dumps(deck)

    def add_card(self, card: Card):
        applicableCard = list(
            filter(lambda x: x.cardCode == card.cardCode, self.cards))
        if applicableCard:
            applicableCard = applicableCard[0]
            applicableCard.count += 1
        else:
            self.cards.append(card)
            self._cards.append(card.cardCode)

    def remove_card(self, card: Card):
        applicableCard = filter(lambda x: x.cardCode == card.cardCode,
                                self.cards)
        if applicableCard:
            applicableCard = applicableCard[0]
            applicableCard.count -= 1
            if applicableCard.count == 0:
                self.cards.remove(applicableCard)
                self._cards.remove(applicableCard.cardCode)

    def __str__(self):
        response = ["Decklist:", "--------------"]

        for card in self.cards:
            response.append(f"({card.cost}) {card.name} x {card.count}")
        return "\n".join(response)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['twisted_fate', 'twisted_fate.api_wrapper', 'twisted_fate.deck_coder']

package_data = \
{'': ['*'], 'twisted_fate': ['data/data/*']}

install_requires = \
['requests>=2.22,<3.0']

setup_kwargs = {
    'name': 'twisted-fate',
    'version': '0.1.4',
    'description': 'An api wrapper around the Legends of Runeterra Client API with deck encoder implementation',
    'long_description': '# twisted_fate\n\n\nA python api wrapper to for riot\'s Legends of Runterra client-api including a port of [Riot Games C# implementation of the deck encoder/decoder](https://github.com/RiotGames/LoRDeckCodes)\n\n# Install\n\n\n```\npip install twisted_fate\n```\n\n## Usage\n\n### Create Deck from deck code\n```python\nfrom twisted_fate import Deck\n\ndraven_deck = Deck.decode("CEBAGAIDCQRSOCQBAQAQYDISDQTCOKBNGQAACAIBAMFQ")\n\n# results\nprint(deck.cards)\n#[\n#    Card(01NX020, Name: Draven, Cost: 3), \n#    Card(01NX035, Name: Draven\'s Biggest Fan, Cost: 1), \n#    Card(01NX039, Name: Vision, Cost: 3), \n#    Card(01PZ001, Name: Rummage, Cost: 1), \n#    Card(01PZ012, Name: Flame Chompers!, Cost: 2), \n#    Card(01PZ013, Name: Augmented Experimenter, Cost: 6), #    Card(01PZ018, Name: Academy Prodigy, Cost: 2), \n#    Card(01PZ028, Name: Jury-Rig, Cost: 1), \n#    Card(01PZ038, Name: Sump Dredger, Cost: 2), \n#    Card(01PZ039, Name: Get Excited!, Cost: 3), \n#    Card(01PZ040, Name: Jinx, Cost: 4), \n#    Card(01PZ045, Name: Zaunite Urchin, Cost: 1), \n#    Card(01PZ052, Name: Mystic Shot, Cost: 2), \n#    Card(01NX011, Name: Whirling Death, Cost: 3)\n# ]\n```\n### Create Deck from cards list, (in the format of a response from the client api)\n```python\nfrom twisted_fate import Deck\n# client api response\ndeck = {\n    "DeckCode":"CEBAGAIDCQRSOCQBAQAQYDISDQTCOKBNGQAACAIBAMFQ",\n    "CardsInDeck": {\n        "01NX020": 3,\n        "01NX035": 3,\n        "01NX039": 3,\n        "01PZ001": 3,\n        "01PZ012": 3,\n        "01PZ013": 3,\n        "01PZ018": 3,\n        "01PZ028": 3,\n        "01PZ038": 3,\n        "01PZ039": 3,\n        "01PZ040": 3,\n        "01PZ045": 3,\n        "01PZ052": 3,\n        "01NX011": 1,\n    },\n}\n\n\ndraven_deck = Deck(cards=deck["CardsInDeck"])\nprint(draven_deck.encode().deck_code)\n# result: CEBAGAIDCQRSOCQBAQAQYDISDQTCOKBNGQAACAIBAMFQ\n\n# or\n\nprint(draven_deck.to_deck_code())\n# result: CEBAGAIDCQRSOCQBAQAQYDISDQTCOKBNGQAACAIBAMFQ\n```\n\n\n## Client API Wrapper\ncomming soon...\n\n\n',
    'author': 'Anthony Keelan',
    'author_email': 'anthony.keelan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Snowcola/twisted_fate',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
import json
from .utils import read_json_file, get_card_set_online
data_dir = Path(__file__).parent.parent / "data/data"
print(data_dir.exists())
print(Path(".").absolute())
print(Path("./data/data/set1-en_us.json").exists())

try:
    cards_files = data_dir.glob("set*.json")
    cards = []
    for f in cards_files:
        f_json = read_json_file(f)
        cards += f_json
except Exception as e:
    print("Could not load card data")


class Card:
    def __init__(self, card=None, **kwargs):
        self.id = kwargs.get("CardID", None)
        self.cardCode = kwargs.get("CardCode", card)
        self.card_set = int(self.cardCode[:2])
        self.count = int(kwargs.get("count", 1))
        self._card_data = self.card_info()

        self.image_path = f"/img/cards/{self.cardCode}.png"
        self.image_path_full = f"./static/img/cards/{self.cardCode}-full.png"

        self.image_online = f"https://raw.githubusercontent.com/pedrofracassi/lor-bundles/master/set{self.card_set}/en_us/img/cards/{self.cardCode}.png"
        self.image_online_full = self.image_online.replace(".png", "-full.png")

    def card_info(self):
        return [card for card in cards if card["cardCode"] == self.cardCode][0]

    def add_copy(self):
        self.count += 1

    def remove_copy(self):
        self.count -= 1

    @property
    def name(self):
        return self._card_data["name"]

    @property
    def isChampion(self):
        return self.superType == "Champion"

    @property
    def description(self):
        return self._card_data["descriptionRaw"]

    @property
    def descriptionFancy(self):
        return self._card_data["description"]

    @property
    def keywords(self):
        return self._card_data["keywords"]

    @property
    def keywordRefs(self):
        return self._card_data["keywordRefs"]

    @property
    def cost(self):
        return self._card_data["cost"]

    @property
    def health(self):
        return self._card_data["health"]

    @property
    def attack(self):
        return self._card_data["attack"]

    @property
    def associatedCardRefs(self):
        return self._card_data["associatedCardRefs"]

    @property
    def associatedCards(self):
        return self._card_data["associatedCards"]

    @property
    def collectible(self):
        return self._card_data["collectible"]

    @property
    def flavorText(self):
        return self._card_data["flavorText"]

    @property
    def rarity(self):
        return self._card_data["rarity"]

    @property
    def rarityRef(self):
        return self._card_data["rarityRef"]

    @property
    def region(self):
        return self._card_data["region"]

    @property
    def spellSpeed(self):
        return self._card_data["spellSpeed"]

    @property
    def spellSpeedRef(self):
        return self._card_data["spellSpeedRef"]

    @property
    def subType(self):
        return self._card_data["subtype"]

    @property
    def superType(self):
        return self._card_data["supertype"]

    @property
    def cardType(self):
        return self._card_data["type"]

    def serialize(self, props=None, as_dict=False):
        if not props:
            props = [
                "name",
                "description",
                "cardCode",
                "keywords",
                "cost",
                "health",
                "attack",
                "flavorText",
                "rarity",
                "rarityRef",
                "spellSpeed",
                "spellSpeedRef",
                "subtype",
                "supertype",
                "type",
            ]

        #s = {k: v for (k, v) in self.__dict__.items() if k in props}
        s = {k: v for (k, v) in self._card_data.items() if k in props}
        s["count"] = self.count
        return s if as_dict else json.dumps(s)

    def __str__(self):
        return f"({self.cost}) {self.name}: {self.description}"

    def __repr__(self):
        return f"Card({self.cardCode}, Name: {self.name}, Cost: {self.cost})"

    def __hash__(self):  # allow using cards as dict keys
        return hash(self.cardCode)



import json
import requests

def read_json_file(json_file):
    with open(json_file,encoding='utf8') as f:
        return json.load(f)


def get_card_set_online(set_num: int, region="en_us"):
    url = f"https://raw.githubusercontent.com/pedrofracassi/lor-bundles/master/set{set_num}/{region}/data/set{set_num}-{region}.json"
    r = requests.get(url)   
    return r.json()


def get_lor_globals(region="en_us"):
    url = f"https://raw.githubusercontent.com/pedrofracassi/lor-bundles/master/core/{region}/data/globals-{region}.json"
    r = requests.get(url)
    return r.json()